// KEDS_count
// Version 3.0
// Rewritten in C++ April 2003 by Nicolas Frisby
// Kansas Event Data System Project
// Philip A. Schrodt
// Dept of Politial Science
// University of Kansas
// Lawrence, KS 66045
// internet:  p-schrodt@ukans.edu

// Revision History
// 1.0 Philip Schrodt in Pascal
// 2.0 Star Straf rewrite in Java
// 2.01 Added rounding to Scale and directory path to command file
// 2.02 Send output files to directory that command file is in
// 3.0 Nicolas Frisby rewrite in C++ (added leap year handling and fixed wildcard functionality)

// C++ version notes: the usage of the tm structure for the dates has some tricky implications mostly
// surrounding the mktime() function; look into it if confused about dates. The majority of the
// uncommented code was basically copied from the Java version with little or no changes; see
// the Java (and perhaps even Pascal) version for comments.

//MAIN PROGRAM
#include "KEDS_count.h"

//UTILITY protos---------------------------
bool Rgetline(ifstream& instream, char* dest, const int MAX_length);
void Converttoupper(char* cp);
void Trimstrncpy(char* dest, char* src, const int MAX_length);
void IndexOf(char* str, int& index, char delim);
void Substring(char* dest, char* src, int begin, int end, const int MAX_length);
bool IsBefore(const tm &curr, const tm &begin);
void FormatDate(char* datestr, const tm &thedate);
bool CheckForBadLeap(const tm &curr);
//UTILITY protos---------------------------

//PROGRAM protos---------------------------
void KEDS_error(char* emsg,char* exmsg = NULL);
void KEDS_warn(char* emsg,char* exmsg = NULL);
//there are many more...
//PROGRAM protos---------------------------

//UTILITY----------------------------------
bool Rgetline(ifstream& instream, char* dest, const int MAX_length)
//reads a line from file up to MAX_length characters
//interprets all newline formats (file must open in ios::bin)
//throws exception upon input overflow
{
	char c = ' ';
	int i = 0;
	char *holder = dest; //keep copy of string start for potential error

	c = instream.peek(); //only peek returns EOF... stream.get() does not

	if ( EOF == c )	//return false if already at end of file
		return false;

	instream.get(c);

	for ( ; EOF != c && i < MAX_length; i++ ) {
		if ( '\n' == c ) {		//catch unix and dos newlines
			break;
		} else if ( '\r' == c ) {	//catch mac newlines and ignore \r part of dos newlines
			c = instream.peek();

			if ( '\n' != c ) {
				break;
			}
		} else {
			*dest++ = c;
		}

		instream.get(c);
	}

	*dest = '\0';

	if ( i == MAX_length ) {
		KEDS_warn("Line length was greater than limit; discarded:", holder);
		return false;
	}

	return true;
}

void Converttoupper(char* cp)
//converts entire string to uppercase
//stops at '\0'
{
	while ( '\0' != *cp) {
		*cp = toupper(*cp);
		++cp;
	}
}

void Trimstrncpy(char* dest, char* src, const int MAX_length)
//copies src to dest, stopping at '\0', ' ', or MAX_length characters
{
	char* csrc = src;
	while ( '\0' != *csrc && ' ' != *csrc )
		++csrc;
	*csrc = '\0';
	strncpy(dest,src,MAX_length);
}

void IndexOf(char* str, int& index, char delim)
//sets index to first occurence of delim or '\0' in str
{
	//find first occurence of delim
	char* pos = strchr(str,delim);

	//check for fail
	if ( NULL == pos ) {
		index = -1;
		return;
	}		

	//ptr arithmetic for range
	index = pos - str;

	//make sure it's positive
	if ( index < 0 )
		index *= -1;
}

void Substring(char* dest, char* src, int begin, int end, const int MAX_length)
//copies src[begin:end] into dest
//null-terminates dest
//stops after MAX_length - 1 characters have been copied
{
	int nchar = end - begin + 1;

	if ( nchar > MAX_length )
		nchar = MAX_length;

	src += begin;

	while ( '\0' != *src && 0 < --nchar ) {
		*dest++ = *src++;
		nchar;
	}

	*dest = '\0';
}

bool IsBefore(const tm &curr, const tm &begin)
//returns a true if the date in time structure curr happens before the date in begin
{
	if ( curr.tm_year > begin.tm_year )
		return false;
	else if ( curr.tm_year < begin.tm_year )
		return true;

	if ( curr.tm_mon > begin.tm_mon )
		return false;
	else if ( curr.tm_mon < begin.tm_mon )
		return true;

	if ( curr.tm_mday > begin.tm_mday )
		return false;
	else if ( curr.tm_mday < begin.tm_mday )
		return true;

	//note: equal dates returns false
	return false;
}

void FormatDate(char* datestr, const tm &thedate)
//creates a date string in datestr that reflects time structure thedate
//format depends upon the value of formatCalendar which is a global var set in the KEDS Count command file
{
	char month[2];
	char year[2];
	char day[2];
	int temp;
	int nyear;
	int i;

	nyear = thedate.tm_year + 1900;
	nyear = nyear - 100*((int)nyear/100);

	temp = nyear/10;
	year[0] = (char)(temp + '0');
	temp = nyear - 10*temp;
	year[1] = (char)(temp + '0');

	temp = (thedate.tm_mon+1)/10;
	month[0] = temp + '0';
	temp = thedate.tm_mon + 1 - 10*temp;
	month[1] = temp + '0';

	temp = thedate.tm_mday/10;
	day[0] = temp + '0';
	temp = thedate.tm_mday - 10*temp;
	day[1] = temp + '0';

	strcpy(datestr,formatCalendar);

	for ( i = 0; i < strlen(datestr); ++i)
	{
		switch ( toupper(datestr[i]) )
		{
			case 'M':
				datestr[i] = month[0];
				datestr[++i] = month[1];
				break;
			case 'Y':
				datestr[i] = year[0];
				datestr[++i] = year[1];
				break;
			case 'D':
				datestr[i] = day[0];
				datestr[++i] = day[1];
				break;
		}
	}

	datestr[i] = '\0';
}

bool CheckForBadLeap(const tm &curr)
//checks for an illegal February 29th
//returns true if the date is legal
//based on Gregorian definition of leap years
{
	if ( 29 == curr.tm_mday && 1 == curr.tm_mon ) //if it's Feb 29
	{
		if ( (curr.tm_year+1900) % 4 != 0 )	//rules for leap year
		{
			return false;
		}
		else if ( (curr.tm_year+1900) % 400 == 0 )
		{
			return true;
		}
		else if ( (curr.tm_year+1900) % 100 == 0 )
		{
			return false;
		}
		else
			return true;
	}
	else
	{
		return true;
	}
}
//UTILITY----------------------------------

//PROGRAM----------------------------------
void KEDS_error(char* emsg, char* exmsg)
//display error and abort
{
	cout << endl << endl << "ERROR--> " << emsg << endl;

	if (NULL != exmsg)
		cout << ">>\t" << exmsg << "\t<<" << endl;

	cout << endl << "Processing unable to continue. Press any key to abort.";

	getch();

	exit(1);
}

void KEDS_warn(char* emsg, char* exmsg)
//potential error, warn user and prompt for continuation
{
	cout << endl << endl << "WARNING--> " << emsg << endl;
	if (NULL != exmsg)
		cout << ">>\t" << exmsg << "\t<<" << endl;
	cout << endl;

	cout << "Processing may be able to continue. Abort? ";

	char c;
	cin.get(c);

	while ( !(toupper(c) == 'Y' || toupper(c) == 'N') ) {
		cout << endl << "Y or N, please. ";
		cin.get(c);
	}

	if ( toupper(c) == 'Y' )
		exit(0);
}

void output_variable_date()
//copy old variable file into new variable file, adding the new column after each line
{
	char priorLine[MAX_VAR_LINE];
	char priorDate[MAX_DATE];
	char currDate[MAX_DATE] = "";
	int datePos;

	FormatDate(currDate,BeginDate);
	*priorLine = '\0';

	for ( int loopvar = 0; loopvar <= maxCount+2; loopvar++)
	{
		if ( !Rgetline(*priorOut[loopvar],priorLine,MAX_VAR_LINE) )
			KEDS_error("Cannot read from prior output file");

		// parse out date and make sure the same, if not error
		IndexOf(priorLine,datePos,TAB);
		Substring(priorDate,priorLine,0,datePos,MAX_DATE);

		if ( 0 != strcmp(currDate,priorDate) )
		{
			char errorLine[MAX_LINE];
			strncpy(errorLine,"Current: ",MAX_LINE);
			strncat(errorLine,currDate,MAX_LINE);
			strncat(errorLine,"does not equal prior: ",MAX_LINE);
			strncat(errorLine,priorDate,MAX_LINE);
			KEDS_error("Date processing problem.",errorLine);
		}

		// write line to new file
		*dyadOut[loopvar] << priorLine;
	}
	// Count
	for ( int loopvar = 0; loopvar <= maxTargets; loopvar++)
	{
		*dyadOut[maxCount+1] << TAB << dayCount[loopvar];
	}
	// Scale
	for ( int loopvar = 0; loopvar <= maxTargets; loopvar++)
	{ //rint to round to two decimal places
		*dyadOut[maxCount+2] << TAB << (float)round(dayScale[loopvar]*100)/100;
	}

	// counts
	for ( int loopC = 0; loopC <= maxCount; loopC++)
	{
		for ( int loopT = 0; loopT <= maxTargets; loopT++)
		{
			*dyadOut[loopC] << TAB << dayCode[loopT][loopC];
		} // loopT
	} // loopC

	for ( int loopvar = 0; loopvar <= maxCount+2; loopvar++)
	{
		*dyadOut[loopvar] << endl;
	}
} // output_variable_date

void output_first_variable_date()
//create first variable output
//this function writes the dates and title to the file
{
	char datestr[MAX_DATE] = "";

	for ( int loopvar = 0; loopvar <= maxCount+2; loopvar++)
	{
		FormatDate(datestr,BeginDate);
		*dyadOut[loopvar] << datestr;
	}
	// Count
	for ( int loopvar = 0; loopvar <= maxTargets; loopvar++)
	{
		*dyadOut[maxCount+1] << TAB << dayCount[loopvar];
	}
	// Scale
	for ( int loopvar = 0; loopvar <= maxTargets; loopvar++)
	{
		*dyadOut[maxCount+2] << TAB << (float)round((dayScale[loopvar])*100)/100; //two decimal places
	}
	// counts
	for ( int loopC = 0; loopC <= maxCount; loopC++)
	{
		for ( int loopT = 0; loopT <= maxTargets; loopT++)
		{
			*dyadOut[loopC] << TAB << dayCode[loopT][loopC];
		} // loopT
	} // loopC
	for ( int loopvar = 0; loopvar <= maxCount+2; loopvar++)
	{
		*dyadOut[loopvar] << endl;
	}
} // output_first_variable_date

void output_dyad_date()
{ // output current date info to output files
	char datestr[MAX_DATE] = "";

	for ( int loopvar = 0; loopvar <= maxTargets; loopvar++)
	{
		FormatDate(datestr,BeginDate);
		*dyadOut[loopvar] << datestr;
		*dyadOut[loopvar] << TAB << dayCount[loopvar] << TAB << (float)round(dayScale[loopvar]*100)/100;

		for ( int loop2 = 0; loop2 <= maxCount; loop2++)
		{
			*dyadOut[loopvar] << TAB << dayCode[loopvar][loop2];
		} // loop2

		*dyadOut[loopvar] << endl;
	} // for
} // ouptut dyad date

void outputCurrDate()
//general output function
//redirects output request to proper function depending upon formatOutput (set in command file)
//prepares next day's variables for use
{
		if ( 0 == strcmp(formatOutput,"DYAD") )
		{
			output_dyad_date();
		}
		else
		{  
			if (firstOutput)
			{
				output_first_variable_date();
			}
		    else
		    {
		    	output_variable_date();
		    }
		}
		
		for ( int loopvar = 0; loopvar <= maxTargets; loopvar++)
		{
			// Clear out day totals for next day 
			dayCount[loopvar] = 0;
			dayScale[loopvar] = 0;
			for ( int loopcnt = 0; loopcnt <= maxCount; loopcnt++)
			{
				dayCode[loopvar][loopcnt] = 0;
			}
		}
} // output curr date

void setEndDate()
{ // set End Date to BeginDate + Frequency
	EndDate.tm_year = BeginDate.tm_year;
	EndDate.tm_mon = BeginDate.tm_mon;
	EndDate.tm_mday = BeginDate.tm_mday;

	if ( 0 == strcmp(formatFreq,"DAILY") )
	{
		EndDate.tm_mday += 1;
	}
	else if ( 0 == strcmp(formatFreq,"WEEKLY") )
	{
		EndDate.tm_mday += 7;
	}
	else if ( 0 == strcmp(formatFreq,"BIWEEKLY") )
	{
		EndDate.tm_mday += 14;
	}
	else if ( 0 == strcmp(formatFreq,"MONTHLY") )
	{
		EndDate.tm_mon += 1;
	}
	else if ( 0 == strcmp(formatFreq,"QUARTERLY") )
	{
		EndDate.tm_mon += 3;
	}
	else
	{
		KEDS_error("Unkown Frequency option in command file");
	}	

	//this call rectifies date:
	//if today BeginDate was 30th and Frequency added, then the day would be 44th
	//mktime converts that into the 14th or 13th of the next month
	mktime ( &EndDate );
	
	return;
}
void setDayOfWeek()
{ // check formatWeek and use that to set weekly to
  // the correct day of week
	int wday = 0;
	int diff = 0;

	//fixes erroneous days and sets the weekday member varible of BeginDate
	mktime( &BeginDate );

	switch ( formatWeek[0] ) {
		case 'S':
			if ( 'A' == formatWeek[1] )
				wday = 6;
			else
				wday = 0;
			break;
		case 'M':
			wday = 1;
			break;
		case 'T':
			if ( 'U' == formatWeek[1] )
				wday = 2;
			else
				wday = 4;
			break;
		case 'W':
			wday = 3;
			break;
		case 'F':
			wday = 5;
			break;
		default:
			KEDS_warn("Bad WEEK option in command file","Using default = SUNDAY");
			return;
	}
		
	//the date is "diff many" days off
	diff = wday - BeginDate.tm_wday;
	
	if ( diff > 0 )
		diff -= 7;

	//shift date
	BeginDate.tm_mday += diff;

	//fix weekday value
	mktime( &BeginDate );
} // setDayOfWeek

bool setCurrDate()
{ // take locDate and put into Curr Date
  // Assume date input in YYMMDD or YYYYMMDD format 
  // 1900 window 28 to 99

	int cmonth = 0;
	int cday   = 1;
	int cyear  = 1999;
	int posm = 0;
	char temp[5];

	if ( strlen(locDate) == 6)
	{
		Substring(temp,locDate,0,2,5);
		cyear = atoi(temp);
		posm = 2;
		if (cyear < 28)
		{
			cyear += 100;
		}
	}
	else if ( strlen(locDate) == 8)
	{
		Substring(temp,locDate,0,4,5);
		cyear = atoi(temp) - 1900;
		posm = 4;
	}

	Substring(temp,locDate,posm,posm+2,5);
	cmonth = atoi(temp); // 1 - 12
	Substring(temp,locDate,posm+2,posm+4,5);
	cday = atoi(temp);

	CurrDate.tm_year = cyear;
	CurrDate.tm_mon = cmonth-1; // 0 - 11
	CurrDate.tm_mday = cday;

	if ( !CheckForBadLeap(CurrDate) )
		return false;

	mktime( &CurrDate );

	// printDate();
	if (firstData)
	{ 
		if ( 0 == strcmp(formatFreq,"MONTHLY") )
		{
			cday = 1;
		}
		if ( 0 == strcmp(formatFreq,"QUARTERLY") )
		{
			cday = 1; 
			if (cmonth < 4)
			{
				cmonth = 1;
			} // Jan
			else if (cmonth < 7)
			{
				cmonth = 4;
			} // April
			else if (cmonth < 10)
				{
					cmonth = 7;
				} // July
			else
			{
				cmonth = 10;
			} // October
		}

		BeginDate.tm_year = cyear;
		BeginDate.tm_mon = cmonth-1;
		BeginDate.tm_mday = cday;
		mktime( &BeginDate );

		if ( 0 == strcmp(formatFreq,"WEEKLY") || 0 == strcmp(formatFreq,"BIWEEKLY") )
		{
			setDayOfWeek();
		}

		firstData = false;
		setEndDate();
	}

	if ( IsBefore(CurrDate,BeginDate) )
	{
		KEDS_warn("Date out of order so skipped ", locDate);
		*locDate = '\0';
	}
	else if ( !IsBefore(CurrDate,EndDate) )
	{ // increment date until in range
		do {
			outputCurrDate();
			BeginDate.tm_year = EndDate.tm_year;
			BeginDate.tm_mon = EndDate.tm_mon;
			BeginDate.tm_mday = EndDate.tm_mday;
			mktime( &BeginDate );
			setEndDate();
		} while ( IsBefore(EndDate,CurrDate) );
	}

	return true;
}

void match_event()
//tabulates results of successfully parsed events
{
	bool match = false;

	if (dspline >= 45)
	{
		cout << endl;
		dspline = 0;
	}

	if (dspcnt >= formatFeedback )
	{
		cout << dspchar;
		dspcnt = 0;
		dspline++;
		dspchar = ' ';
	}

	//matches source
	if ( dWildcard ) //wildcard in source
		match = ( strstr(locSrc,dSource) == locSrc );
	else			// no wildcard in source
		match = ( 0 == strcmp(locSrc,dSource) );

	if ( !match )
	{
		if ( ' ' == dspchar )
	    {
	    	dspchar = '-';
	    }
		dspcnt++;
	}
	else 
	{
		for (int loop = 0; loop <= maxTargets; loop++)
		{
			if ( tAll || 0 == strcmp(dTarget[loop],locTar) || (dTWildcard[loop] && strstr(locTar,dTarget[loop]) == locTar) )
			{
				if ( ' ' == dspchar || '-' == dspchar || '*' == dspchar )
				{
					dspchar = '+';
				}

				dspcnt++;
				dayCount[loop]++;
				for (int sclLoop = 0; sclLoop <= maxCodeValue; sclLoop++)
				{
					if ( 0 == strcmp(locEvt,code[sclLoop]) )
					{
						dayScale[loop] += value[sclLoop];
					}
				} // add scale value
				for (int cntLoop = 0; cntLoop <= maxCount; cntLoop++)
				{
					if ( 0 == strcmp(countcodes[cntLoop],locEvt) || (wildcard[cntLoop] && strstr(locEvt,countcodes[cntLoop]) == locEvt) )
					{
						dayCode[loop][cntLoop]++;
					}
				} // add code
			} // if match
			else
			{
				if ( ' ' == dspchar || '-' == dspchar )
				{
				dspchar = '*';
				}

				dspcnt++;
			} // Source match only
		} // for loop thru targets
	} // else Source matched
} // match_event

bool parse_event(char* dline)
{ // divide dataLine into Date, Src, Tar, Evt
  // based on columns in format

	int tabpos, column;
	char eventLine[MAX_LINE];
	char temp[MAX_LINE];
	char eventItem[MAX_INPUT];

	strncpy(eventLine,dline,MAX_LINE);
	column = 0;
	*locDate = '\0';
	*locSrc = '\0';
	*locTar = '\0';
	*locEvt = '\0';

	while ( '\0' != *eventLine )
	{
		IndexOf(eventLine,tabpos,TAB);	//tab delimited input

		if (tabpos <= 0)
		{
			Trimstrncpy(eventItem,eventLine,MAX_INPUT);
			*eventLine = '\0';
		}
		else
		{
			Trimstrncpy(temp,eventLine,MAX_LINE);
			Substring(eventItem,temp,0,tabpos,MAX_INPUT);
			strncpy(temp,eventLine,MAX_LINE);
			Substring(eventLine,temp,tabpos+1,strlen(temp),MAX_LINE);
		}
		column++;

		if (column == formatDate)
		{
			strncpy(locDate,eventItem,MAX_INPUT);

			if ( !setCurrDate() )
				*locDate = '\0';
		}
		else if (column == formatSource)
		{
			strncpy(locSrc,eventItem,MAX_INPUT);
		}
		else if (column == formatTarget)
		{
			strncpy(locTar,eventItem,MAX_INPUT);
		}
		else if (column == formatEvent)
		{
			strncpy(locEvt,eventItem,MAX_INPUT);
		}
	}  // while
	
	//check for incomplete or invalid event data
	if (( '\0' == *locDate ) || ( '\0' == *locSrc ) || ( '\0' == *locTar ) || ( '\0' == *locEvt ))
	{
		KEDS_warn("Discarded bad event data line: ", dline);
		dspline = 0;
		return false;
	} 
	else
	{
		return true;
	}
} // parse event

void close_variable_files()
{
	char priorfilename[MAX_FILENAME];
	char newfilename[MAX_FILENAME];
	char dummy[2] = {'\0','\0'};

	for ( int loopvar = 0; loopvar <= maxCount+2; loopvar++)
	{
		dyadOut[loopvar]->close();
		delete dyadOut[loopvar]; //return dynamically allocated ofstreams

		if (!firstOutput)
		{
			priorOut[loopvar]->close();
			if ( priorOut[loopvar]->fail() )
			{
				delete priorOut[loopvar]; //return dynamically allocated ifstreams
				KEDS_warn("Problem closing output files" );
				return;
			}
			delete priorOut[loopvar]; //return dynamically allocated ifstreams
		}
	} // for looop

	// rename each output file from $name to name and delete temporary name
	if (!firstOutput)
	{
		strcpy(priorfilename,formatPrefix);
		strcat(priorfilename,"N");
		strcpy(newfilename,"$");
		strcat(newfilename,priorfilename);
		remove(priorfilename);
		rename(newfilename,priorfilename);

		strcpy(priorfilename,formatPrefix);
		strcat(priorfilename,"SCALE");
		strcpy(newfilename,"$");
		strcat(newfilename,priorfilename);
		remove(priorfilename);
		rename(newfilename,priorfilename);

		// Counts
		for (int loop2 = 0; loop2 <= maxCount; loop2++)
		{ 
			dummy[0] = loop2+'0';
			strcpy(priorfilename,formatPrefix);
			strcat(priorfilename,"CNT");
			strcat(priorfilename,dummy);
			strcpy(newfilename,"$");
			strcat(newfilename,priorfilename);

			remove(priorfilename);
			rename(newfilename,priorfilename);
		}
	} //if
	firstOutput = false;
} // close_variable_files

void open_variable_files()
//opens the output file for variable output
{
	char priorfilename[MAX_FILENAME];
	char fullopen[MAX_KC_PATH];
	char newfilename[MAX_FILENAME];
	char dummy[2] = {'\0','\0'};
	char priorLine[MAX_VAR_LINE];

	for ( int loopvar = 0; loopvar <= maxTargets; loopvar++)
	{
		dayCount[loopvar] = 0;
		dayScale[loopvar] = 0;
		for ( int loopcnt = 0; loopcnt <= maxCount; loopcnt++)
		{
			dayCode[loopvar][loopcnt] = 0;
		}
	}

	if (!firstOutput)
	{ // Open prior version as priorOut
	// Open new one as $

		// Count
		strcpy(priorfilename,formatPrefix);
		strcat(priorfilename,"N");
		strcpy(newfilename,"$");
		strcat(newfilename,priorfilename);
		strcpy(fullopen,inDir);
		strcat(fullopen,newfilename);
		dyadOut[maxCount+1] = new ofstream (fullopen); //dynamically allocated output file streams.
		strcpy(fullopen,inDir);
		strcat(fullopen,priorfilename);
		priorOut[maxCount+1] = new ifstream (fullopen); //dynamically allocated input file streams.

		if ( dyadOut[maxCount+1]->fail() )
		{
			KEDS_error("Problem opening ", newfilename );
			return;
		}

		// Scale
		strcpy(priorfilename,formatPrefix);
		strcat(priorfilename,"SCALE");
		strcpy(newfilename,"$");
		strcat(newfilename,priorfilename);
		strcpy(fullopen,inDir);
		strcat(fullopen,newfilename);
		dyadOut[maxCount+2] = new ofstream (fullopen); //dynamically allocated output file streams.
		strcpy(fullopen,inDir);
		strcat(fullopen,priorfilename);
		priorOut[maxCount+2] = new ifstream (fullopen); //dynamically allocated input file streams.

		if ( dyadOut[maxCount+2]->fail() )
		{
			KEDS_error("Problem opening ", newfilename );
			return;
		}

		// Counts
		for (int loop2 = 0; loop2 <= maxCount; loop2++)
		{
			dummy[0] = loop2+'0';
			strcpy(priorfilename,formatPrefix);
			strcat(priorfilename,"CNT");
			strcat(priorfilename,dummy);
			strcpy(newfilename,"$");
			strcat(newfilename,priorfilename);

			strcpy(fullopen,inDir);
			strcat(fullopen,newfilename);
			dyadOut[loop2] = new ofstream (fullopen); //dynamically allocated output file streams.
			strcpy(fullopen,inDir);
			strcat(fullopen,priorfilename);
			priorOut[loop2] = new ifstream (fullopen); //dynamically allocated output file streams.
		}

		for (int loop4 = 0; loop4 <= maxCount+2; loop4++)
		{
			priorOut[loop4]->getline(priorLine,MAX_VAR_LINE);
			*dyadOut[loop4] << priorLine << endl;
			priorOut[loop4]->getline(priorLine,MAX_VAR_LINE);
			*dyadOut[loop4] << priorLine;
			for (int loop5 = 0; loop5 <= maxTargets; loop5++)
			{
				*dyadOut[loop4] << TAB << dSource;
				if ( dWildcard ) 
				{
					*dyadOut[loop4] << '*';
				}
				*dyadOut[loop4] << '>' << dTarget[loop5];
				if ( !tAll && dTWildcard[loop5] )
				{
					*dyadOut[loop4] << '*';
				}
			}
			*dyadOut[loop4] << endl;
		}
	} //if

	if (firstOutput)
	{ // else rename at back and keep open forreading }
		for (int loop2 = 0; loop2 <= maxCount; loop2++)
		{
			dummy[0] = loop2+'0';
			strcpy(newfilename,formatPrefix);
			strcat(newfilename,"CNT");
			strcat(newfilename,dummy);

			strcpy(fullopen,inDir);
			strcat(fullopen,newfilename);
			dyadOut[loop2] = new ofstream (fullopen); //dynamically allocated output file streams.
		}

		strcpy(newfilename,formatPrefix);
		strcat(newfilename,"N");
		strcpy(fullopen,inDir);
		strcat(fullopen,newfilename);
		dyadOut[maxCount+1] = new ofstream (fullopen); //dynamically allocated output file streams.

		if ( dyadOut[maxCount+1]->fail() )
		{
			KEDS_error("Problem opening ", newfilename );
			return;
		}

		strcpy(newfilename,formatPrefix);
		strcat(newfilename,"SCALE");
		strcpy(fullopen,inDir);
		strcat(fullopen,newfilename);
		dyadOut[maxCount+2] = new ofstream (fullopen); //dynamically allocated output file streams.

		if ( dyadOut[maxCount+2]->fail() )
		{
			KEDS_error("Problem opening ", newfilename );
			return;
		}

		*dyadOut[maxCount+1] << "COUNT" << endl;
		*dyadOut[maxCount+2] << "SCALE" << endl;
		for (int loop3 = 0; loop3 <= maxCount; loop3++)
		{
			*dyadOut[loop3] << "CODE " << countcodes[loop3];
			if (wildcard[loop3])
			{
				*dyadOut[loop3] << '*';
			}
			*dyadOut[loop3] << endl;
		}
		for (int loop4 = 0; loop4 <= maxCount+2; loop4++)
		{
			*dyadOut[loop4] << "Date";
			for (int loop5 = 0; loop5 <= maxTargets; loop5++)
			{
				*dyadOut[loop4] << TAB << dSource;
				if ( dWildcard ) *dyadOut[loop4] << '*';
				*dyadOut[loop4] << '>' << dTarget[loop5];
				if ( !tAll && dTWildcard[loop5] )
				{
					*dyadOut[loop4] << '*';
				}
			}
			*dyadOut[loop4] << endl;
		}
	} // if
} // open_variable_files


void close_dyad_files()
{
	for ( int loopvar = 0; loopvar <= maxTargets; loopvar++)
	{
		dyadOut[loopvar]->close();
		delete dyadOut[loopvar];	//return dynamically allocated output file streams
	} // for looop
} // close_dyad_files

void open_dyad_files()
//converts *s to 'x's and >s to '2's for legal and appropriate filenames
{
	char fullopen[MAX_KC_PATH];
	char filename[MAX_FILENAME];
	char* star;

	for ( int loopvar = 0; loopvar <= maxTargets; loopvar++)
	{
		dayCount[loopvar] = 0;
		dayScale[loopvar] = 0;

		for ( int loopcnt = 0; loopcnt <= maxCount; loopcnt ++)
		{
			dayCode[loopvar][loopcnt] = 0;
		}

		strncpy(filename,formatPrefix,MAX_FILENAME);
		strncat(filename,dSource,MAX_FILENAME);
		if ( dWildcard )
		{
			strncat(filename,"*",MAX_FILENAME);
		}
		strncat(filename,"2",MAX_FILENAME);
		strncat(filename,dTarget[loopvar],MAX_FILENAME);
		if ( !tAll && dTWildcard[loopvar] )
		{
			strncat(filename,"*",MAX_FILENAME);
		}
		
		star = strchr(filename,'*');

		while ( NULL != star )
		{	//'*' is not legal for filenames, so replace it with 'x'
			*star = 'x';
			star = strchr(filename,'*');
		}			

		strcpy(fullopen,inDir);
		strcat(fullopen,filename);
		dyadOut[loopvar] = new ofstream (fullopen); //dynamically allocated output file streams.
		if ( dyadOut[loopvar]->fail() )
			{
			  KEDS_error("Problem opening file:", filename);
			  return;
			}
			
		*dyadOut[loopvar] << "Date" << TAB << "N" << TAB << "Scale";

		for ( int loop2 = 0; loop2 <= maxCount; loop2++)
		{
			*dyadOut[loopvar] << TAB << countcodes[loop2];
			if ( wildcard[loop2] )
			{
				*dyadOut[loopvar] << "*";
			}
		} // loop2

		*dyadOut[loopvar] << endl;
	} // for
} // open_dyad_files

// ******************************************************************
// *               Process Dyad                                     *
// ******************************************************************

void process_dyad()
{
	char fullopen[MAX_KC_PATH];
	char dataLine[MAX_EVENT_LINE] = "";
	bool processedAtLeastOne = false;

	firstData = true;  //  Needed to set begin Date

	if ( 0 == strcmp("DYAD",formatOutput) )
	{
		open_dyad_files();
	}
	else
	{
		open_variable_files();
	}

	cout << endl << "============= PROCESSING SOURCE =============" << endl;
	cout << dSource;
	if ( dWildcard ) cout << '*';
	cout << endl;

	dspcnt = 0;
	dspline = 0;
	dspchar = ' ';

	for ( int fileloop = 0; fileloop <= maxFiles; fileloop++)
	{  // go thru the data files  files[]
		ifstream datain;
		
		//open next file
		strcpy(fullopen,inDir);
		strcat(fullopen,files[fileloop]);
		datain.open(fullopen);
		if ( datain.fail() ) {
			KEDS_warn("Cannot read from data file");
			datain.clear();
			continue;
		}
		processedAtLeastOne = true;

		while ( Rgetline(datain,dataLine,MAX_EVENT_LINE) )
		{  //  process the event file line by line
			if ( parse_event(dataLine) )
			{
				match_event();
			}
		}  // while 
		datain.close();
	} // file loop

	if ( !processedAtLeastOne )
	{
			KEDS_error("No input event files found.");
	}

	outputCurrDate();

	if ( 0 == strcmp("DYAD",formatOutput) )
	{
		close_dyad_files();
	}
	else
	{
		close_variable_files();
	}
}  // end of KEDS_process_data

// ******************************************************************
// *               Read in Dyads                                    *
// ******************************************************************



void read_dyad(char* dyadLine)
// read in dyad line <source>/<target1><target2>..<targetn>
{  
	int pos1,pos2,pos3;
	char targetLine[MAX_LINE] = "";
	char tempLine[MAX_LINE] = "";

	// re-initialize dyad variables
	*dSource = '\0';
	dWildcard = false;
	maxTargets = -1;
	tAll = false;

	cout << endl << endl << "=============== READING DYAD ================" << endl;
	cout << dyadLine << endl;

	IndexOf(dyadLine,pos1,'/');
	IndexOf(dyadLine,pos2,'*');

	//parse for dyad source
	if ( ( -1 == pos2 ) || (pos2 > pos1)) //IndexOf sets pos1/2 to -1 if the delim is not found
	{	//for dyad sets with no wildcards or only wildcards in the targets
		Substring(dSource,dyadLine,0,pos1,MAX_DYAD);
		dWildcard = false;
	}
	else
	{	//for dyad sets with wildcard in the source
		Substring(dSource,dyadLine,0,pos2,MAX_DYAD);
		dWildcard = true;
	}

	//copy all of target string into targetLine
	Substring(targetLine,dyadLine,pos1+1,strlen(dyadLine),MAX_LINE);

	//parse targetLine
	while ( '\0' != *targetLine )
	{
		IndexOf(targetLine,pos3,' ');
		IndexOf(targetLine,pos2,'*');
		if (maxTargets == TARGETSMAX)
		{
			KEDS_warn("To many entries in Target section. Ignored: ",targetLine);
		}
		else
		{
			maxTargets++;
			if ( 0 == pos2 ) //current target is "***" --> since '*' was first char
			{
				strncpy(dTarget[maxTargets],"***",MAX_DYAD);
				dTWildcard[maxTargets] = true;
				tAll = true;
			} // All
			else if (( -1 == pos2 ) || (pos2 > pos3)) //current target does not have wildcard
			{ 
				if (pos3 > 0) //multiple targets left in targetLine
				{
					Substring(tempLine,targetLine,0,pos3,MAX_LINE);
					strncpy(dTarget[maxTargets],tempLine,MAX_DYAD);
				}
				else //only one target (left) in targetLine
				{  
					strncpy(dTarget[maxTargets],targetLine,MAX_DYAD);
				}
				dTWildcard[maxTargets] = false;
			}
			else //target has wildcard
			{ 
				Substring(tempLine,targetLine,0,pos2,MAX_LINE);
				strncpy(dTarget[maxTargets],tempLine,MAX_DYAD);
				dTWildcard[maxTargets] = true;
			}

			if (pos3 > 0) //targetLine is not empty
			{
				Substring(tempLine,targetLine,pos3+1,strlen(targetLine),MAX_LINE);
				strncpy(targetLine,tempLine,MAX_LINE);
			}
			else //no more targets
			{
				*targetLine = '\0';
			}	
		} // Not TARGETSMAX
	} // while anything left in targetLine
} // read_dyad

void count_section(char* cntLine)
{ // read in codes to be counted, may contain * wildcard
	int pos;

	IndexOf(cntLine,pos,'*');

	maxCount++;

	//IndexOf sets pos = -1 if '*' is not found
	if ( -1 == pos )
	{ 
		Trimstrncpy(countcodes[maxCount],cntLine,MAX_COUNT);
		wildcard[maxCount] = false;
	}
	else
	{ 
		Substring(countcodes[maxCount],cntLine,0,pos,MAX_COUNT);
		wildcard[maxCount] = true; //remember this code includes a wildcard
	}
	return;
}

void manual_files_section()
 // if no files in command file then prompt user for files
 // and store in KC.FileList.text
{
	char fullopen[MAX_KC_PATH];
  	char inputFileName[MAX_INPUT] = "";
	ofstream output;

	cout << "Input data files " << endl;

	do 
	{
		cout << "Select Data File: (strike enter when finished)" << endl;
		cin.getline(inputFileName,MAX_INPUT);
		if ( '\0' != *inputFileName )
		{
			cout << "Data file: " << inputFileName << endl;
			maxFiles++;
			strncpy(files[maxFiles],inputFileName,MAX_FILENAME);
		}
	} while (( '\0' != *inputFileName ) || (maxFiles < 0));

	strcpy(fullopen,inDir);
	strcat(fullopen,"KC.FileList.text");
	output.open(fullopen,ios::out);
	if ( output.fail() ) {
		KEDS_error("Problem opening KC.FileList.text");
		return;
	}

	for ( int loopvar = 0; loopvar <= maxFiles; loopvar++)
	{
		output << files[loopvar] << endl;
	}

	output.close();
	cout << "The file names have been written to the file" << " 'KC.FileList.text' " << endl << "You can copy these into your command file for future use " << endl;
	return;
}

void files_section(char* fileLine)
{ // read in file names
	if ( FILESMAX == maxFiles )
	{
		KEDS_warn("To many entries in FILE section. Ignored: ", fileLine);
	}
	else
	{
		maxFiles++;
		Trimstrncpy(files[maxFiles],fileLine,MAX_FILENAME);
	}
	return;
}	

void format_section(char* fmtLine)
{ // read in line from -FORMAT section of command file
	char tempCal[MAX_CAL];
	char* tempFor;

	int pos;
	IndexOf(fmtLine,pos,'=');
	pos += 1;
	
	//catch bad lines
	if ( -1 == pos ) {
		KEDS_warn("Bad input line in FORMAT section. Ignored: ",fmtLine);
		return;
	}
	
	//logical assert: pos is the index of character after '='

	if ( 'P' == fmtLine[0] )  // PREFIX
	{
		strncpy(formatPrefix,fmtLine+pos,MAX_PREF);
	}
	else if ( 'D' == fmtLine[0] )  // DATE
	{
		formatDate = atoi(fmtLine+pos);
	}
	else if ( 'S' == fmtLine[0] )  // SOURCE
	{
		formatSource = atoi(fmtLine+pos);
	}
	else if ( 'T' == fmtLine[0] )  // TARGET
	{
		formatTarget = atoi(fmtLine+pos);
	}
	else if ( 'E' == fmtLine[0] )  // EVENT
	{
		formatEvent = atoi(fmtLine+pos);
	}
	else if ( 'F' == fmtLine[0] && 'R' == fmtLine[1] )  // FREQ
	{
		strncpy(formatFreq,fmtLine+pos,MAX_FREQ);
	}
	else if ( 'F' == fmtLine[0] && 'E' == fmtLine[1] )  // Feedback
	{
		formatFeedback = atoi(fmtLine+pos);
	}
	else if ( 'C' == fmtLine[0] )  // CALENDAR
	{
		Substring(tempCal,fmtLine,pos,strlen(fmtLine),MAX_CAL);
		Converttoupper(tempCal);
		strncpy(formatCalendar,tempCal,MAX_CAL);
	}
	else if ( 'W' == fmtLine[0] ) // WEEKDAY
	{
		Trimstrncpy(formatWeek,fmtLine+pos,MAX_WEEK);
		Converttoupper(formatWeek);
	}
	else if ( 'O' == fmtLine[0] ) // OUTPUT
	{
		tempFor = fmtLine + pos; // skip past '='
		
		Converttoupper(tempFor);

		if ( 'D' == tempFor[0] && 'Y' == tempFor[1] )
		{
			strcpy(formatOutput,"DYAD");
		}
		else if ( 'V' == tempFor[0] && 'A' == tempFor[1] )
		{
			strcpy(formatOutput,"VARIABLE");
		}
		
		cout << "Processing as " << formatOutput << endl;
	}
	else
	{
		KEDS_error("Unknown section in FORMAT section",fmtLine);
	}
	return;
}

void scales_section(char* sclLine)
{ // read in <code>:[value] event codes and their scale
	char  	codestr[MAX_CODE] = "";
	char 	scs[MAX_VAL] = "";
	float  	scn;
	int 	pos1,pos2,pos3;

	IndexOf(sclLine,pos1,':');
	IndexOf(sclLine,pos2,'[');
	IndexOf(sclLine,pos3,']');

	if ( -1 == pos1 )
	{
		*codestr = '\0';
		KEDS_warn("Invalid code in SCALE section. Ignored: ", sclLine);
	}
	else
	{
		Substring(codestr,sclLine,0,pos1,MAX_CODE);
	}
	if ( -1 == pos2 || -1 == pos3 )
	{
		scn = 0;
	}
	else
	{
		Substring(scs,sclLine,pos2+1,pos3,MAX_VAL); 
		scn = atof(scs);
	}
	if ( CODEVALUEMAX == maxCodeValue )
	{
		KEDS_warn("To many entries in SCALE section. Ignored: ",sclLine);
		*codestr = '\0';
	}
	if ( '\0' != *codestr )
	{
		maxCodeValue++;
		strncpy(code[maxCodeValue],codestr,MAX_CODE);
		value[maxCodeValue] = scn;
	}
	return;
}

char determine_section(char* secLine)
{  // determine which section you are at in the command file
	char temp;
	temp = ' ';

	if (strcmp(secLine,"-FORMAT") == 0) 
	{
		cout << "STATUS: Reading formats" << endl;
		temp = 'F';
	}
	// documentation says SCALES but example is CODES so allow either
	else if (strcmp(secLine,"-SCALES") == 0) 
	{
		cout << "STATUS: Reading event codes and scales" << endl;
		temp = 'S';
	}
	else if (strcmp(secLine,"-CODES")  == 0) 
	{
		cout << "STATUS: Reading event codes and scales" << endl;
		temp = 'S';
	}
	else if (strcmp(secLine,"-FILES")  == 0) 
	{
		cout << "STATUS: Reading files" << endl;
		temp = 'L';
	}
	else if (strcmp(secLine,"-COUNT")  == 0) 
	{
		cout << "STATUS: Reading codes to count" << endl;
		temp = 'C';
	}
	else if (strcmp(secLine,"-DYADS")  == 0) 
	{
		cout << "STATUS: Processing dyads" << endl;
		cout << "Feedback every " << formatFeedback << " entries" << endl;
		temp = 'D';
	}
	else if (strcmp(secLine,"-FINISH") == 0) 
	{
		temp = 'E';
	}
	else
	{
		KEDS_error("Unknown section in command file",secLine);
	} // unknown section
	return temp;
}

void KEDS_read_command_file()
{  	// read in the command file and display progress on screen
	int slashpos = -1;
	char thisLine[MAX_LINE];
	char inputFilePath[MAX_KC_PATH];
	char inputFileName[MAX_FILENAME];
	char* slash = NULL;
	char* tempSlash = NULL;
	char section = ' ';
	bool programcomplete = false;
	char holder;

	ifstream inputFile;

	do
	{
	    inputFile.clear();
	    
	    cout << "Select Command File:" << endl;
		cin.getline(inputFilePath,MAX_KC_PATH);
		
		holder = *(inputFilePath+4);
		*(inputFilePath+4) = '\0';

		if ( 0 == stricmp(inputFilePath,"quit") ) {
			cout << "Exiting program." << endl;
			exit(0);
		}

		*(inputFilePath+4) = holder;
		
		inputFile.open(inputFilePath,ios::binary);
	} while ( inputFile.fail() );

	slash = strrchr(inputFilePath,'\\');	//escaped backslash
	tempSlash = strrchr(inputFilePath,'/');

	//ASSUMPTION!!! the file system format is controlled by the last occurence of a slash (forward or backwards) in the file path.
	if ( NULL != tempSlash || NULL != slash )
	{
		if ( tempSlash > slash )
			slash = tempSlash;

		strncpy(inputFileName,(slash+1),MAX_FILENAME);
		*(slash+1) = '\0';
		strncpy(inDir,inputFilePath,MAX_DIR);

		cout << "Working Directory: " << inDir << endl;
		cout << "Command file: " << inputFileName << endl;
	}
	else
	{
		strncpy(inputFileName,inputFilePath,MAX_FILENAME);
		*inDir = '\0';
		cout << "Working Directory: ." << endl;
		cout << "Command file: " << inputFileName << endl;
	}

	while ( !programcomplete && Rgetline(inputFile,thisLine,MAX_LINE) )
	{  //  process the command file line by line
		if ( '-' == thisLine[0] )  //  if a new section
		{
       		section = determine_section(thisLine);
		} // new section if statement
		else
		{
			switch(section)
			{
				case 'F' : // -FORMAT
					format_section(thisLine);
				break;
				case 'S' : // -SCALES or -CODES
					scales_section(thisLine);
				break;  
				case 'L' : // -FILES
					files_section(thisLine);
				break;  
				case 'C' : // -COUNT
					if (maxFiles == -1)
					{ // no files in command file, prompt user for files
						manual_files_section();
					}
					count_section(thisLine);
				break;  
				case 'D' : // -DYADS
					if (maxFiles == -1)
					{ // no files in command file, prompt user for files
						manual_files_section();
					}
					read_dyad(thisLine);
					process_dyad();
				break;  
				case 'E' : // -FINISH
					cout << endl << endl << "STATUS: Complete" << endl;
					programcomplete = true;
				break;  
			} // switch statement 
		}  //  data for a section (else)
	}  // while 
	inputFile.close();
}  //  end of KEDS_read_command_file

int main( void )
{
	//default values
	strcpy(formatPrefix,"KCOUNT.");
	strcpy(formatFreq,"MONTHLY");
	strcpy(formatCalendar,"YYMMDD");
	strcpy(formatOutput,"DYAD");

	KEDS_read_command_file();

	//beep and wait for key stroke
	Beep(5000,500);
	getch();

	return 0;
}
//PROGRAM----------------------------------
